# -*- coding: utf-8 -*-
import time
import warnings

import numpy as np
import torch
from ase import Atoms
from torch_geometric.loader import DataLoader as DataLoader_pyg

from mattersim.datasets.utils.convertor import GraphConvertor


def build_dataloader(
    atoms: list[Atoms] = None,
    energies: list[float] = None,
    forces: list[np.ndarray] = None,
    stresses: list[np.ndarray] = None,
    cutoff: float = 5.0,
    threebody_cutoff: float = 4.0,
    batch_size: int = 64,
    model_type: str = "m3gnet",
    shuffle=False,
    only_inference: bool = False,
    num_workers: int = 0,
    pin_memory: bool = False,
    multiprocessing: int = 0,
    multithreading: int = 0,
    dataset=None,
    finetune_task_label: list = None,
    **kwargs,
):
    """
    Build a dataloader given a list of atoms
        - atoms : a list of atoms in ase format
        - energies, forces and stresses are necessary for training
            - energies : a list of energy (float) with unit eV
            - forces : a list of nx3 force matrix (np.ndarray) with unit eV/Å,
                where n is the number of atom in each structure.
            - stresses : a list of 3x3 stress matrix (np.ndarray) with unit GPa
        - only_inference : if True, energies, forces and stresses will be ignored
        - num_workers : number of workers for dataloader
        - pin_memory : if True, the datasets will be stored in GPU or CPU memory
        - pin_memory_device : the device for pin_memory
        - dataset : the dataset object for the dataloader
                    only used for graphormer and geomformer
    """

    convertor = GraphConvertor(model_type, cutoff, True, threebody_cutoff)

    preprocessed_data = []

    if dataset is None:
        if not only_inference:
            assert (
                energies is not None
            ), "energies must be provided if only_inference is False"
        if stresses is not None:
            assert np.array(stresses[0]).shape == (
                3,
                3,
            ), "stresses must be a list of 3x3 matrices"

        length = len(atoms)
        if energies is None:
            energies = [None] * length
        if forces is None:
            forces = [None] * length
        if stresses is None:
            stresses = [None] * length

    if model_type == "m3gnet":
        if multiprocessing == 0 and multithreading == 0:
            # start = time.time()
            for graph, energy, force, stress in zip(atoms, energies, forces, stresses):
                graph = convertor.convert(graph.copy(), energy, force, stress, **kwargs)
                if graph is not None:
                    preprocessed_data.append(graph)
            # print("Data preprocessing time: {:.2f} s".format(time.time() - start))
        elif multithreading > 0 and multiprocessing == 0:
            from multiprocessing.pool import ThreadPool

            warnings.warn("multithreading is experimental")
            warnings.warn("it may not be faster than single thread due to GIL.")
            print("Using multithreading with {} threads".format(multithreading))
            start = time.time()
            pool = ThreadPool(processes=multithreading)
            preprocessed_data = pool.starmap(
                convertor.convert, zip(atoms, energies, forces, stresses)
            )
            pool.close()
            print("Time elapsed: {:.2f} s".format(time.time() - start))
        elif multiprocessing > 0 and multithreading == 0:
            import multiprocessing as mp

            warnings.warn("multiprocessing is experimental.")
            print("Using multiprocessing with {} workers".format(multiprocessing))
            # torch.multiprocessing.set_sharing_strategy('file_system')
            start = time.time()
            pool = mp.Pool(multiprocessing)
            results = []
            for i in range(multiprocessing):
                left = int(i * length / multiprocessing)
                right = int((i + 1) * length / multiprocessing)
                results.append(
                    pool.apply_async(multiprocess_data, args=(atoms[left:right], 1))
                )
            pool.close()
            pool.join()
            for result in results:
                graph = result.get()
                if graph is not None:
                    preprocessed_data.extend(graph)
            print("Time for multiprocessing: {:.2f} s".format(time.time() - start))
        else:
            raise NotImplementedError

        return DataLoader_pyg(
            preprocessed_data,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=pin_memory,
        )

    elif model_type == "graphormer" or model_type == "geomformer":
        raise NotImplementedError


def multiprocess_data(atoms: list[Atoms], number):
    convertor = GraphConvertor()
    result = []
    for graph in atoms:
        graph = convertor.convert(
            graph,
            graph.get_potential_energy(),
            graph.get_forces(),
            graph.get_stress(voigt=False) * 160.2,
        )
        if graph is not None:
            result.append(graph)
    return result



def pad_1d_unsqueeze(x, padlen):
    x = x + 1  # pad id = 0
    xlen = x.size(0)
    if xlen < padlen:
        new_x = x.new_zeros([padlen], dtype=x.dtype)
        new_x[:xlen] = x
        x = new_x
    return x.unsqueeze(0)


def pad_2d_unsqueeze(x, padlen):
    x = x + 1  # pad id = 0
    xlen, xdim = x.size()
    if xlen < padlen:
        new_x = x.new_zeros([padlen, xdim], dtype=x.dtype)
        new_x[:xlen, :] = x
        x = new_x
    return x.unsqueeze(0)


@torch.jit.script
def mask_after_k_persample(n_sample: int, n_len: int, persample_k: torch.Tensor):
    assert persample_k.shape[0] == n_sample
    assert persample_k.max() <= n_len
    device = persample_k.device
    mask = torch.zeros([n_sample, n_len + 1], device=device)
    mask[torch.arange(n_sample, device=device), persample_k] = 1
    mask = mask.cumsum(dim=1)[:, :-1]
    return mask.type(torch.bool)


def auto_cell(cell, cutoff=10.0):
    # find max value in x, y, z direction
    max_x = max(int(cutoff / torch.min(torch.abs(cell[:, 0, 0]))), 1)
    max_y = max(int(cutoff / torch.min(torch.abs(cell[:, 1, 1]))), 1)
    max_z = max(int(cutoff / torch.min(torch.abs(cell[:, 2, 2]))), 1)
    # loop
    cells = []
    for i in range(-max_x, max_x + 1):
        for j in range(-max_y, max_y + 1):
            for k in range(-max_z, max_z + 1):
                if i == 0 and j == 0 and k == 0:
                    continue
                cells.append([i, j, k])
    return cells


def cell_expand(pos, atoms, cell, cutoff=10.0):
    batch_size, max_num_atoms = pos.size()[:2]
    cells = auto_cell(cell, cutoff)
    cell_tensor = (
        torch.tensor(cells, device=pos.device)
        .to(cell.dtype)
        .unsqueeze(0)
        .expand(batch_size, -1, -1)
    )  # batch_size, n_cell, 3
    offset = torch.bmm(cell_tensor, cell)  # B x n_cell x 3
    expand_pos = pos.unsqueeze(1) + offset.unsqueeze(2)  # B x n_cell x T x 3
    expand_pos = expand_pos.view(batch_size, -1, 3)  # B x (n_cell x T) x 3
    expand_dist = torch.norm(
        pos.unsqueeze(2) - expand_pos.unsqueeze(1), p=2, dim=-1
    )  # B x T x (8 x T)
    expand_mask = expand_dist < cutoff  # B x T x (8 x T)
    expand_mask = torch.masked_fill(expand_mask, atoms.eq(0).unsqueeze(-1), False)
    expand_mask = (torch.sum(expand_mask, dim=1) > 0) & (
        ~(atoms.eq(0).repeat(1, len(cells)))
    )  # B x (8 x T)
    expand_len = torch.sum(expand_mask, dim=-1)
    max_expand_len = torch.max(expand_len)
    outcell_index = torch.zeros(
        [batch_size, max_expand_len], dtype=torch.long, device=pos.device
    )
    expand_pos_compressed = torch.zeros(
        [batch_size, max_expand_len, 3], dtype=pos.dtype, device=pos.device
    )
    outcell_all_index = torch.arange(
        max_num_atoms, dtype=torch.long, device=pos.device
    ).repeat(len(cells))
    for i in range(batch_size):
        outcell_index[i, : expand_len[i]] = outcell_all_index[expand_mask[i]]
        expand_pos_compressed[i, : expand_len[i], :] = expand_pos[i, expand_mask[i], :]
    return (
        expand_pos_compressed,
        expand_len,
        outcell_index,
        mask_after_k_persample(batch_size, max_expand_len, expand_len),
    )


def pad_spatial_pos_unsqueeze(x, padlen):
    x = x + 1
    xlen = x.size(0)
    if xlen < padlen:
        new_x = x.new_zeros([padlen, padlen], dtype=x.dtype)
        new_x[:xlen, :xlen] = x
        x = new_x
    return x.unsqueeze(0)


@torch.jit.script
def convert_to_single_emb(x, offset: int = 512):
    feature_num = x.size(1) if len(x.size()) > 1 else 1
    feature_offset = 1 + torch.arange(0, feature_num * offset, offset, dtype=torch.long)
    x = x + feature_offset
    return x


class BatchedDataDataset(torch.utils.data.Dataset):
    # class BatchedDataDataset(torch.utils.data.IterableDataset):
    def __init__(
        self,
        dataset,
        max_node=512,
        infer=False,
    ):
        super().__init__()
        self.dataset = dataset
        self.max_node = max_node

        self.infer = infer

    def __getitem__(self, index):
        item = self.dataset[int(index)]
        return item

    def __len__(self):
        return len(self.dataset)

    def collate(self, samples):
        return collator_ft(
            samples,
            max_node=self.max_node,
            use_pbc=True,
        )


def pad_pos_unsqueeze(x, padlen):
    xlen, xdim = x.size()
    if xlen < padlen:
        new_x = x.new_zeros([padlen, xdim], dtype=x.dtype)
        new_x[:xlen, :] = x
        x = new_x
    return x.unsqueeze(0)


def collator_ft(items, max_node=512, use_pbc=True):
    original_len = len(items)
    items = [item for item in items if item is not None and item.x.size(0) <= max_node]
    filtered_len = len(items)
    if filtered_len < original_len:
        pass
        # print("warning: molecules with atoms more than %d are filtered" % max_node)
    pos = None
    max_node_num = max(item.x.size(0) for item in items if item is not None)
    forces = None
    stress = None
    total_energy = None

    if hasattr(items[0], "pos") and items[0].pos is not None:
        poses = [item.pos - item.pos.mean(dim=0, keepdim=True) for item in items]
        # poses = [item.pos for item in items]
        pos = torch.cat([pad_pos_unsqueeze(i, max_node_num) for i in poses])
    if hasattr(items[0], "forces") and items[0].forces is not None:
        forcess = [item.forces for item in items]
        forces = torch.cat([pad_pos_unsqueeze(i, max_node_num) for i in forcess])
    if hasattr(items[0], "stress") and items[0].stress is not None:
        stress = torch.cat([item.stress.unsqueeze(0) for item in items], dim=0)
    if hasattr(items[0], "total_energy") and items[0].cell is not None:
        total_energy = torch.cat([item.total_energy for item in items])

    items = [
        (
            item.idx,
            item.x,
            item.y,
            (item.pbc if hasattr(item, "pbc") else torch.tensor([False, False, False]))
            if use_pbc
            else None,
            (item.cell if hasattr(item, "cell") else torch.zeros([3, 3]))
            if use_pbc
            else None,
            (int(item.num_atoms) if hasattr(item, "num_atoms") else item.x.size()[0]),
        )
        for item in items
    ]
    (
        idxs,
        xs,
        ys,
        pbcs,
        cells,
        natoms,
    ) = zip(*items)

    y = torch.cat(ys)
    x = torch.cat([pad_2d_unsqueeze(i, max_node_num) for i in xs])

    pbc = torch.cat([i.unsqueeze(0) for i in pbcs], dim=0) if use_pbc else None
    cell = torch.cat([i.unsqueeze(0) for i in cells], dim=0) if use_pbc else None
    natoms = torch.tensor(natoms) if use_pbc else None
    node_type_edge = None
    return dict(
        idx=torch.LongTensor(idxs),
        x=x,
        y=y,
        pos=pos + 1e-5,
        pbc=pbc,
        cell=cell,
        natoms=natoms,
        total_energy=total_energy,
        forces=forces,
        stress=stress,
        node_type_edge=node_type_edge,
    )



# -*- coding: utf-8 -*-
import os

from ase import Atoms
from ase.io import read as ase_read
from mp_api.client import MPRester
from pymatgen.core.structure import Structure
from pymatgen.io.ase import AseAtomsAdaptor


class AtomsAdaptor(object):
    """
    This class is used to read different structures type
    and transform it to ASE Atoms object.
    """

    def __init__(self):
        pass

    @classmethod
    def from_ase_atoms(cls, atoms: Atoms):
        """
        Get Atoms from Atoms.

        Args:
            atoms (Atoms): ASE Atoms object.
        """
        if not isinstance(atoms, Atoms):
            raise TypeError("Input must be ASE Atoms object.")
        return atoms

    @classmethod
    def from_pymatgen_structure(cls, structure: Structure):
        """
        Get Atoms from Structure.

        Args:
            structure (Structure): pymatgen Structure object.
        """
        if not isinstance(structure, Structure):
            raise TypeError("Input must be pymatgen Structure object")
        return AseAtomsAdaptor.get_atoms(structure, msonable=False)

    @classmethod
    def from_mp_id(cls, mp_id: str, api_key: str = None):
        """
        Get Atoms from mp-id.

        mp_id (str): mp_id for materials.
        api_key (str, optional): api_key to access Material Projects.
            If not provided, try to extract it from environment variables.
        """
        mp_api_key = api_key or os.getenv("MP_API_KEY")
        if not mp_api_key:
            raise ValueError(
                "An MP API key is required to fetch data from"
                " Materials Project, but was not found in the"
                " environment variables or provided."
            )
        with MPRester(mp_api_key) as m:
            structure = m.get_structure_by_material_id(mp_id)
            return AseAtomsAdaptor.get_atoms(structure, msonable=False)

    @classmethod
    def from_file(cls, filename: str, format: str = None):
        """
        Get Atoms from file.

        filename (str): file name which contains structures.
        format (str, optional): file format. If None, will automately
            guess.
        """
        if not os.path.exists(filename):
            raise FileNotFoundError(f"File {filename} not found.")

        if format:
            atoms_list = ase_read(filename, index=":", format=format)
        else:
            try:
                atoms_list = ase_read(filename, index=":")
            except Exception as e:
                raise ValueError(f"Can not automately guess the file format: {e}")

        return atoms_list



# -*- coding: utf-8 -*-
from typing import Union

import torch.nn as nn


class LinearLayer(nn.Module):
    def __init__(
        self,
        in_dim,
        out_dim,
        bias=True,
    ):
        super().__init__()
        self.linear = nn.Linear(in_dim, out_dim, bias=bias)

    def forward(self, x):
        return self.linear(x)


class SigmoidLayer(nn.Module):
    def __init__(
        self,
        in_dim,
        out_dim,
        bias=True,
    ):
        super().__init__()
        self.linear = nn.Linear(in_dim, out_dim, bias=bias)
        self.sigmoid = nn.Sigmoid()

    def forward(
        self,
        x,
    ):
        return self.sigmoid(self.linear(x))


class SwishLayer(nn.Module):
    def __init__(
        self,
        in_dim,
        out_dim,
        bias=True,
    ):
        super().__init__()
        self.linear = nn.Linear(in_dim, out_dim, bias=bias)
        self.sigmoid = nn.Sigmoid()

    def forward(
        self,
        x,
    ):
        x = self.linear(x)
        return x * self.sigmoid(x)


class ReLULayer(nn.Module):
    def __init__(
        self,
        in_dim,
        out_dim,
        bias=True,
    ):
        super().__init__()
        self.linear = nn.Linear(in_dim, out_dim, bias=bias)
        self.relu = nn.ReLU()

    def forward(
        self,
        x,
    ):
        return self.relu(self.linear(x))


class GatedMLP(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dims: list,
        activation: Union[list[Union[str, None]], str] = "swish",
        use_bias: bool = True,
    ):
        super().__init__()
        input_dim = in_dim
        if isinstance(activation, str) or activation is None:
            activation = [activation] * len(out_dims)
        else:
            assert len(activation) == len(
                out_dims
            ), "activation and out_dims must have the same length"
        module_list_g = []
        for i in range(len(out_dims)):
            if activation[i] == "swish":
                module_list_g.append(  # noqa: E501
                    SwishLayer(input_dim, out_dims[i], bias=use_bias)
                )
            elif activation[i] == "sigmoid":
                module_list_g.append(
                    SigmoidLayer(input_dim, out_dims[i], bias=use_bias)
                )
            elif activation[i] is None:
                module_list_g.append(  # noqa: E501
                    LinearLayer(input_dim, out_dims[i], bias=use_bias)
                )
            input_dim = out_dims[i]
        module_list_sigma = []
        activation[-1] = "sigmoid"
        input_dim = in_dim
        for i in range(len(out_dims)):
            if activation[i] == "swish":
                module_list_sigma.append(
                    SwishLayer(input_dim, out_dims[i], bias=use_bias)
                )
            elif activation[i] == "sigmoid":
                module_list_sigma.append(
                    SigmoidLayer(input_dim, out_dims[i], bias=use_bias)
                )
            elif activation[i] is None:
                module_list_sigma.append(
                    LinearLayer(input_dim, out_dims[i], bias=use_bias)
                )
            else:
                raise NotImplementedError
            input_dim = out_dims[i]
        self.g = nn.Sequential(*module_list_g)
        self.sigma = nn.Sequential(*module_list_sigma)

    def forward(
        self,
        x,
    ):
        return self.g(x) * self.sigma(x)


class MLP(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dims: list,
        activation: Union[list[Union[str, None]], str, None] = "swish",
        use_bias: bool = True,
    ):
        super().__init__()
        input_dim = in_dim
        if isinstance(activation, str) or activation is None:
            activation = [activation] * len(out_dims)
        else:
            assert len(activation) == len(
                out_dims
            ), "activation and out_dims must have the same length"
        module_list = []
        for i in range(len(out_dims)):
            if activation[i] == "swish":
                module_list.append(
                    SwishLayer(input_dim, out_dims[i], bias=use_bias)  # noqa: E501
                )
            elif activation[i] == "sigmoid":
                module_list.append(
                    SigmoidLayer(input_dim, out_dims[i], bias=use_bias)  # noqa: E501
                )
            elif activation[i] is None:
                module_list.append(
                    LinearLayer(input_dim, out_dims[i], bias=use_bias)  # noqa: E501
                )
            else:
                raise NotImplementedError
            input_dim = out_dims[i]
        self.mlp = nn.Sequential(*module_list)

    def forward(
        self,
        x,
    ):
        return self.mlp(x)



# -*- coding: utf-8 -*-
import warnings
from typing import Optional, Tuple

import ase
import numpy as np
import torch
from ase import Atoms
from pymatgen.optimization.neighbors import find_points_in_spheres
from torch_geometric.data import Data

from .threebody_indices import compute_threebody as _compute_threebody

# Ensure the warning is only shown once
warnings.filterwarnings("once", category=UserWarning)

"""
Supported Properties:
    - "num_nodes"(set by default)  ## int
    - "num_edges"(set by default)  ## int
    - "num_atoms"                  ## int
    - "num_bonds"                  ## int
    - "atom_attr"                  ## tensor [num_atoms,atom_attr_dim=1]
    - "atom_pos"                   ## tensor [num_atoms,3]
    - "edge_length"                ## tensor [num_edges,1]
    - "edge_vector"                ## tensor [num_edges,3]
    - "edge_index"                 ## tensor [2,num_edges]
    - "three_body_indices"         ## tensor [num_three_body,2]
    - "num_three_body"              ## int
    - "num_triple_ij"              ## tensor [num_edges,1]
    - "num_triple_i"               ## tensor [num_atoms,1]
    - "num_triple_s"               ## tensor [1,1]
    - "theta_jik"                  ## tensor [num_three_body,1]
    - "triple_edge_length"         ## tensor [num_three_body,1]
    - "phi"                        ## tensor [num_three_body,1]
    - "energy"                     ## float
    - "forces"                     ## tensor [num_atoms,3]
    - "stress"                     ## tensor [3,3]
"""

"""
Computing various graph based operations (M3GNet)
"""


def compute_threebody_indices(
    bond_atom_indices: np.array,
    bond_length: np.array,
    n_atoms: int,
    atomic_number: np.array,
    threebody_cutoff: Optional[float] = None,
):
    """
    Given a graph without threebody indices, add the threebody indices
    according to a threebody cutoff radius
    Args:
        bond_atom_indices: np.array, [n_atoms, 2]
        bond_length: np.array, [n_atoms]
        n_atoms: int
        atomic_number: np.array, [n_atoms]
        threebody_cutoff: float, threebody cutoff radius

    Returns:
        triple_bond_indices, n_triple_ij, n_triple_i, n_triple_s

    """
    n_atoms = np.array(n_atoms).reshape(1)
    atomic_number = atomic_number.reshape(-1, 1)
    n_bond = bond_atom_indices.shape[0]
    if n_bond > 0 and threebody_cutoff is not None:
        valid_three_body = bond_length <= threebody_cutoff
        ij_reverse_map = np.where(valid_three_body)[0]
        original_index = np.arange(n_bond)[valid_three_body]
        bond_atom_indices = bond_atom_indices[valid_three_body, :]
    else:
        ij_reverse_map = None
        original_index = np.arange(n_bond)

    if bond_atom_indices.shape[0] > 0:
        bond_indices, n_triple_ij, n_triple_i, n_triple_s = _compute_threebody(
            np.ascontiguousarray(bond_atom_indices, dtype="int32"),
            np.array(n_atoms, dtype="int32"),
        )
        if ij_reverse_map is not None:
            n_triple_ij_ = np.zeros(shape=(n_bond,), dtype="int32")
            n_triple_ij_[ij_reverse_map] = n_triple_ij
            n_triple_ij = n_triple_ij_
        bond_indices = original_index[bond_indices]
        bond_indices = np.array(bond_indices, dtype="int32")
    else:
        bond_indices = np.reshape(np.array([], dtype="int32"), [-1, 2])
        if n_bond == 0:
            n_triple_ij = np.array([], dtype="int32")
        else:
            n_triple_ij = np.array([0] * n_bond, dtype="int32")
        n_triple_i = np.array([0] * len(atomic_number), dtype="int32")
        n_triple_s = np.array([0], dtype="int32")
    return bond_indices, n_triple_ij, n_triple_i, n_triple_s


def get_fixed_radius_bonding(
    structure: ase.Atoms,
    cutoff: float = 5.0,
    numerical_tol: float = 1e-8,
    pbc: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Get graph representations from structure within cutoff
    Args:
        structure (pymatgen Structure or molecule)
        cutoff (float): cutoff radius
        numerical_tol (float): numerical tolerance

    Returns:
        center_indices, neighbor_indices, images, distances
    """
    pbc_ = np.array(structure.pbc, dtype=int)

    lattice_matrix = np.ascontiguousarray(structure.cell[:], dtype=float)  # noqa: E501

    cart_coords = np.ascontiguousarray(
        np.array(structure.positions), dtype=float
    )  # noqa: E501
    r = float(cutoff)

    (
        center_indices,
        neighbor_indices,
        images,
        distances,
    ) = find_points_in_spheres(  # noqa: E501
        cart_coords,
        cart_coords,
        r=r,
        pbc=pbc_,
        lattice=lattice_matrix,
        tol=numerical_tol,
    )
    center_indices = center_indices.astype(np.int64)
    neighbor_indices = neighbor_indices.astype(np.int64)
    images = images.astype(np.int64)
    distances = distances.astype(float)
    exclude_self = (center_indices != neighbor_indices) | (
        distances > numerical_tol
    )  # noqa: E501
    return (
        center_indices[exclude_self],
        neighbor_indices[exclude_self],
        images[exclude_self],
        distances[exclude_self],
    )


class GraphConvertor:
    """
    Convert ase.Atoms to Graph
    """

    default_properties = ["num_nodes", "num_edges"]

    def __init__(
        self,
        model_type: str = "m3gnet",
        twobody_cutoff: float = 5.0,
        has_threebody: bool = True,
        threebody_cutoff: float = 4.0,
    ):
        self.model_type = model_type
        self.twobody_cutoff = twobody_cutoff
        self.threebody_cutoff = threebody_cutoff
        self.has_threebody = has_threebody

    def convert(
        self,
        atoms: Atoms,
        energy=None,
        forces=None,
        stress=None,
        pbc=True,
        **kwargs,
    ):
        """
        Convert the structure into graph
        Args:
            pbc: bool, whether to use periodic boundary condition, default True
        """
        # normalize the structure
        if isinstance(atoms, Atoms):
            pbc_ = np.array(atoms.pbc, dtype=int)
            if np.all(pbc_ < 0.01) or not pbc:
                min_x = np.min(atoms.positions[:, 0])
                min_y = np.min(atoms.positions[:, 1])
                min_z = np.min(atoms.positions[:, 2])
                max_x = np.max(atoms.positions[:, 0])
                max_y = np.max(atoms.positions[:, 1])
                max_z = np.max(atoms.positions[:, 2])      
                x_len = (max_x - min_x) + max(
                    self.twobody_cutoff, self.threebody_cutoff
                ) * 5
                y_len = (max_y - min_y) + max(
                    self.twobody_cutoff, self.threebody_cutoff
                ) * 5
                z_len = (max_z - min_z) + max(
                    self.twobody_cutoff, self.threebody_cutoff
                ) * 5
                max_len = max(x_len, y_len, z_len)
                x_len = y_len = z_len = max_len
                lattice_matrix = np.eye(3) * max_len
                pbc_ = np.array([1, 1, 1], dtype=int)
                warnings.warn(
                    "No PBC detected, using a large supercell with "
                    f"size {x_len}x{y_len}x{z_len} Angstrom**3",
                    UserWarning,
                )
               
                atoms.set_cell(lattice_matrix)
                atoms.set_pbc(pbc_)
            else:
                if np.all(abs(atoms.cell) < 1e-5):
                    raise ValueError("Cell vectors are too small")
        else:
            raise ValueError("structure type not supported")

        scaled_pos = atoms.get_scaled_positions()
        scaled_pos = np.mod(scaled_pos, 1)
        atoms.set_scaled_positions(scaled_pos)
        args = {}
        if self.model_type == "m3gnet":
            args["num_atoms"] = len(atoms)
            args["num_nodes"] = len(atoms)
            args["atom_attr"] = torch.FloatTensor(
                atoms.get_atomic_numbers()
            ).unsqueeze(  # noqa: E501
                -1
            )
            args["atom_pos"] = torch.FloatTensor(atoms.get_positions())
            args["cell"] = torch.FloatTensor(np.array(atoms.cell)).unsqueeze(0)
            (
                sent_index,
                receive_index,
                shift_vectors,
                distances,
            ) = get_fixed_radius_bonding(atoms, self.twobody_cutoff, pbc=pbc)
            args["num_bonds"] = len(sent_index)
            args["edge_index"] = torch.from_numpy(
                np.array([sent_index, receive_index])
            )  # noqa: E501
            args["pbc_offsets"] = torch.FloatTensor(shift_vectors)
            if self.has_threebody:
                (
                    triple_bond_index,
                    n_triple_ij,
                    n_triple_i,
                    n_triple_s,
                ) = compute_threebody_indices(
                    bond_atom_indices=args["edge_index"]
                    .numpy()
                    .transpose(1, 0),  # noqa: E501
                    bond_length=distances,
                    n_atoms=atoms.positions.shape[0],
                    atomic_number=atoms.get_atomic_numbers(),
                    threebody_cutoff=self.threebody_cutoff,
                )
                args["three_body_indices"] = torch.from_numpy(
                    triple_bond_index
                ).to(  # noqa: E501
                    torch.long
                )  # [num_three_body,2]
                args["num_three_body"] = args["three_body_indices"].shape[0]
                args["num_triple_ij"] = (
                    torch.from_numpy(n_triple_ij).to(torch.long).unsqueeze(-1)
                )
            else:
                args["three_body_indices"] = None
                args["num_three_body"] = None
                args["num_triple_ij"] = None
            if energy is not None:
                args["energy"] = torch.FloatTensor([energy])
            if forces is not None:
                args["forces"] = torch.FloatTensor(forces)
            if stress is not None:
                args["stress"] = torch.FloatTensor(stress).unsqueeze(0)
            return Data(**args)

        elif self.model_type == "graphormer":
            raise NotImplementedError
        else:
            raise NotImplementedError(
                "model type {} not implemented".format(self.model_type)
            )



# -*- coding: utf-8 -*-
from .angle_encoding import SphericalBasisLayer
from .edge_encoding import SmoothBesselBasis
from .layers import MLP, GatedMLP, LinearLayer, SwishLayer
from .message_passing import AtomLayer, EdgeLayer, MainBlock

__all__ = [
    "SphericalBasisLayer",
    "SmoothBesselBasis",
    "GatedMLP",
    "MLP",
    "LinearLayer",
    "SwishLayer",
    "AtomLayer",
    "EdgeLayer",
    "MainBlock",
]



# -*- coding: utf-8 -*-
"""
Atomic scaling module. Used for predicting extensive properties.
"""

from typing import Optional, Union

import numpy as np
import torch
import torch.nn as nn
from ase import Atoms
from torch_runstats.scatter import scatter_mean

from mattersim.datasets.utils.regressor import solver

DATA_INDEX = {
    "total_energy": 0,
    "forces": 2,
    "per_atom_energy": 1,
    "per_species_energy": 0,
}


class AtomScaling(nn.Module):
    """
    Atomic extensive property rescaling module
    """

    def __init__(
        self,
        atoms: list[Atoms] = None,
        total_energy: list[float] = None,
        forces: list[np.ndarray] = None,
        atomic_numbers: list[np.ndarray] = None,
        num_atoms: list[float] = None,
        max_z: int = 94,
        scale_key: str = None,
        shift_key: str = None,
        init_scale: Union[torch.Tensor, float] = None,
        init_shift: Union[torch.Tensor, float] = None,
        trainable_scale: bool = False,
        trainable_shift: bool = False,
        verbose: bool = False,
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
        **kwargs,
    ):
        """
        Args:
            forces: a list of atomic forces (np.ndarray) in each graph
            max_z: (int) maximum atomic number
                - if scale_key or shift_key is specified,
                  max_z should be equal to the maximum atomic_number.
            scale_key: valid options are:
                - total_energy_std
                - per_atom_energy_std
                - per_species_energy_std
                - forces_rms
                - per_species_forces_rms (default)
            shift_key: valid options are:
                - total_energy_mean
                - per_atom_energy_mean
                - per_species_energy_mean :
                  default option is gaussian regression (NequIP)
                - per_species_energy_mean_linear_reg :
                  an alternative choice is linear regression (M3GNet)
            init_scale (torch.Tensor or float)
            init_shift (torch.Tensor or float)
        """
        super().__init__()

        self.max_z = max_z
        self.device = device

        # === Data preprocessing ===
        if scale_key or shift_key:
            total_energy = torch.from_numpy(np.array(total_energy))
            forces = (
                torch.from_numpy(np.concatenate(forces, axis=0))
                if forces is not None
                else None
            )
            if atomic_numbers is None:
                atomic_numbers = [atom.get_atomic_numbers() for atom in atoms]
            atomic_numbers = (
                torch.from_numpy(np.concatenate(atomic_numbers, axis=0))
                .squeeze(-1)
                .long()
            )  # (num_atoms,)
            # assert max_z == atomic_numbers.max().item(),
            # "max_z should be equal to the maximum atomic_number"
            if num_atoms is None:
                num_atoms = [  # noqa: E501
                    atom.positions.shape[0] for atom in atoms
                ]  # (N_GRAPHS, )
            num_atoms = torch.from_numpy(np.array(num_atoms))
            per_atom_energy = total_energy / num_atoms
            data_list = [total_energy, per_atom_energy, forces]

            assert (
                num_atoms.size()[0] == total_energy.size()[0]
            ), "num_atoms and total_energy should have the same size, "
            f"but got {num_atoms.size()[0]} and {total_energy.size()[0]}"
            if forces is not None:
                assert (
                    forces.size()[0] == atomic_numbers.size()[0]
                ), "forces and atomic_numbers should have the same length, "
                f"but got {forces.size()[0]} and {atomic_numbers.size()[0]}"

            # === Calculate the scaling factors ===
            if (
                scale_key == "per_species_energy_std"
                and shift_key == "per_species_energy_mean"
                and init_shift is None
                and init_scale is None
            ):
                # Using gaussian regression two times
                # to get the shift and scale is potentially unstable
                init_shift, init_scale = self.get_gaussian_statistics(
                    atomic_numbers, num_atoms, total_energy
                )
            else:
                if shift_key and init_shift is None:
                    init_shift = self.get_statistics(
                        shift_key, max_z, data_list, atomic_numbers, num_atoms
                    )
                if scale_key and init_scale is None:
                    init_scale = self.get_statistics(
                        scale_key, max_z, data_list, atomic_numbers, num_atoms
                    )

        # === initial values are given ===
        if init_scale is None:
            init_scale = torch.ones(max_z + 1)
        elif isinstance(init_scale, float):
            init_scale = torch.tensor(init_scale).repeat(max_z + 1)
        else:
            assert init_scale.size()[0] == max_z + 1

        if init_shift is None:
            init_shift = torch.zeros(max_z + 1)
        elif isinstance(init_shift, float):
            init_shift = torch.tensor(init_shift).repeat(max_z + 1)
        else:
            assert init_shift.size()[0] == max_z + 1

        init_shift = init_shift.float()
        init_scale = init_scale.float()
        if trainable_scale is True:
            self.scale = torch.nn.Parameter(init_scale)
        else:
            self.register_buffer("scale", init_scale)

        if trainable_shift is True:
            self.shift = torch.nn.Parameter(init_shift)
        else:
            self.register_buffer("shift", init_shift)

        if verbose is True:
            print("Current scale: ", init_scale)
            print("Current shift: ", init_shift)

        self.to(device)

    def transform(
        self, atomic_energies: torch.Tensor, atomic_numbers: torch.Tensor
    ) -> torch.Tensor:
        """
        Take the origin values from model and get the transformed values
        """
        curr_shift = self.shift[atomic_numbers]
        curr_scale = self.scale[atomic_numbers]
        normalized_energies = curr_scale * atomic_energies + curr_shift
        return normalized_energies

    def inverse_transform(
        self, atomic_energies: torch.Tensor, atomic_numbers: torch.Tensor
    ) -> torch.Tensor:
        """
        Take the transformed values and get the original values
        """
        curr_shift = self.shift[atomic_numbers]
        curr_scale = self.scale[atomic_numbers]
        unnormalized_energies = (atomic_energies - curr_shift) / curr_scale
        return unnormalized_energies

    def forward(
        self, atomic_energies: torch.Tensor, atomic_numbers: torch.Tensor
    ) -> torch.Tensor:
        """
        Atomic_energies and atomic_numbers should have the same size
        """
        return self.transform(atomic_energies, atomic_numbers)

    def get_statistics(
        self, key, max_z, data_list, atomic_numbers, num_atoms
    ) -> torch.Tensor:
        """
        Valid key:
            scale_key: valid options are:
                - total_energy_mean
                - per_atom_energy_mean
                - per_species_energy_mean
                - per_species_energy_mean_linear_reg :
                  an alternative choice is linear regression
            shift_key: valid options are:
                - total_energy_std
                - per_atom_energy_std
                - per_species_energy_std
                - forces_rms
                - per_species_forces_rms
        """
        data = None
        for data_key in DATA_INDEX:
            if data_key in key:
                data = data_list[DATA_INDEX[data_key]]
        assert data is not None

        statistics = None
        if "mean" in key:
            if "per_species" in key:
                n_atoms = torch.repeat_interleave(repeats=num_atoms)
                if "linear_reg" in key:
                    features = bincount(
                        atomic_numbers, n_atoms, minlength=self.max_z + 1
                    ).numpy()
                    # print(features[0], features.shape)
                    data = data.numpy()
                    assert features.ndim == 2  # [batch, n_type]
                    features = features[
                        (features > 0).any(axis=1)
                    ]  # deal with non-contiguous batch indexes
                    statistics = np.linalg.pinv(features.T.dot(features)).dot(
                        features.T.dot(data)
                    )
                    statistics = torch.from_numpy(statistics)
                else:
                    N = bincount(
                        atomic_numbers,
                        num_atoms,
                        minlength=self.max_z + 1,  # noqa: E501
                    )
                    assert N.ndim == 2  # [batch, n_type]
                    # deal with non-contiguous batch indexes
                    N = N[(N > 0).any(dim=1)]
                    N = N.type(torch.get_default_dtype())
                    statistics, _ = solver(
                        N, data, regressor="NormalizedGaussianProcess"
                    )
            else:
                statistics = torch.mean(data).item()
        elif "std" in key:
            if "per_species" in key:
                print(
                    "Warning: calculating per_species_energy_std for "
                    "full periodic table systems is risky, please use "
                    "per_species_forces_rms instead."
                )
                n_atoms = torch.repeat_interleave(repeats=num_atoms)
                N = bincount(atomic_numbers, n_atoms, minlength=self.max_z + 1)
                assert N.ndim == 2  # [batch, n_type]
                # deal with non-contiguous batch indexes
                N = N[(N > 0).any(dim=1)]
                N = N.type(torch.get_default_dtype())
                _, statistics = solver(  # noqa: E501
                    N, data, regressor="NormalizedGaussianProcess"
                )
            else:
                statistics = torch.std(data).item()
        elif "rms" in key:
            if "per_species" in key:
                square = scatter_mean(
                    data.square(), atomic_numbers, dim=0, dim_size=max_z + 1
                )
                statistics = square.mean(axis=-1)
            else:
                statistics = torch.sqrt(torch.mean(data.square())).item()

        if isinstance(statistics, torch.Tensor) is not True:
            statistics = torch.tensor(statistics).repeat(max_z + 1)

        assert statistics.size()[0] == max_z + 1

        return statistics

    def get_gaussian_statistics(
        self,
        atomic_numbers: torch.Tensor,
        num_atoms: torch.Tensor,
        total_energy: torch.Tensor,
    ):
        """
        Get the gaussian process mean and variance
        """
        n_atoms = torch.repeat_interleave(repeats=num_atoms)
        N = bincount(atomic_numbers, n_atoms, minlength=self.max_z + 1)
        assert N.ndim == 2  # [batch, n_type]
        N = N[(N > 0).any(dim=1)]  # deal with non-contiguous batch indexes
        N = N.type(torch.get_default_dtype())
        mean, std = solver(  # noqa: E501
            N, total_energy, regressor="NormalizedGaussianProcess"
        )
        assert mean.size()[0] == self.max_z + 1
        assert std.size()[0] == self.max_z + 1
        return mean, std


def bincount(
    input: torch.Tensor,
    batch: Optional[torch.Tensor] = None,
    minlength: int = 0,  # noqa: E501
):
    assert input.ndim == 1
    if batch is None:
        return torch.bincount(input, minlength=minlength)
    else:
        assert batch.shape == input.shape

        length = input.max().item() + 1
        if minlength == 0:
            minlength = length
        if length > minlength:
            raise ValueError(
                f"minlength {minlength} too small for input "
                f"with integers up to and including {length}"  # noqa: E501
            )

        # Flatten indexes
        # Make each "class" in input into a per-input class.
        input_ = input + batch * minlength

        num_batch = batch.max() + 1

        return torch.bincount(input_, minlength=minlength * num_batch).reshape(
            num_batch, minlength
        )



# -*- coding: utf-8 -*-
import os
import sys

from recommonmark.parser import CommonMarkParser  # noqa: F401

sys.path.append(os.path.abspath("../src"))

# -- Project information -----------------------------------------------------
project = "MatterSim"
copyright = "2024"
author = "Microsoft Corporation"
release = "1.0.0"

# -- General configuration ---------------------------------------------------
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx.ext.mathjax",
    "sphinx_copybutton",
    "recommonmark",
    "nbsphinx",
]

source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
html_theme = "sphinx_book_theme"
html_theme_options = {
    "repository_url": "https://github.com/microsoft/mattersim",
    "repository_provider": "github",
    "use_repository_button": True,
}

# colorful codes
# pygments_style = 'sphinx'  # or any other Pygments style you prefer
# pygments_style = 'solarized-dark'  # or any other Pygments style you prefer

html_static_path = ["_static"]

html_logo = "_static/mattersim-with-name.png"

# -- Options for nbsphinx ----------------------------------------------------
nbsphinx_allow_errors = True
nbsphinx_execute = "never"



import os
import re
import uuid
from collections import defaultdict
from typing import List

import pandas as pd
from ase import Atoms
from ase.io import read
from loguru import logger
from pymatgen.io.ase import AseAtomsAdaptor

from mattersim.applications.moldyn import MolecularDynamics


def moldyn(
    atoms_list: List[Atoms],
    *,
    temperature: float = 300,
    timestep: float = 1,
    steps: int = 1000,
    ensemble: str = "nvt_nose_hoover",
    logfile: str = "-",
    loginterval: int = 10,
    trajectory: str = None,
    taut: float = None,
    work_dir: str = str(uuid.uuid4()),
    save_csv: str = "results.csv.gz",
    **kwargs,
) -> dict:
    if len(atoms_list) != 1:
        raise ValueError("molecular dynamics workflow currently only supports one structure")

    moldyn_results = defaultdict(list)

    for atoms in atoms_list:
        # check if the atoms object has non-zero values in the lower triangle
        # of the cell. If so, the cell will be rotated and permuted to upper
        # triangular form. This is to avoid numerical issues in the MD simulation.
        print(atoms.cell.array)
        if any(atoms.cell.array[2, 0:2]) or atoms.cell.array[1, 0] != 0:
            logger.warning(
                "The lower triangle of the cell is not zero. "
                "The cell will be rotated and permuted to upper triangular form."
            )

            # The following code is from the PR
            # https://gitlab.com/ase/ase/-/merge_requests/3277.
            # It will be removed once the PR is merged.
            # This part of the codes rotates the cell and permutes the axes
            # such that the cell will be in upper triangular form.

            from ase.build import make_supercell

            _calc = atoms.calc
            logger.info(f"Initial cell: {atoms.cell.array}")

            atoms.set_cell(atoms.cell.standard_form()[0], scale_atoms=True)

            # Permute a and c axes
            atoms = make_supercell(atoms, [[0, 0, 1], [0, 1, 0], [1, 0, 0]])

            atoms.rotate(90, "y", rotate_cell=True)

            # set the lower triangle of the cell to be exactly zero
            # to avoid numerical issues
            atoms.cell.array[1, 0] = 0
            atoms.cell.array[2, 0] = 0
            atoms.cell.array[2, 1] = 0

            logger.info(f"Cell after rotation/permutation: {atoms.cell.array}")
            atoms.calc = _calc

        if not os.path.exists(work_dir):
            os.makedirs(work_dir)

        if os.path.exists(os.path.join(work_dir, logfile)):
            os.remove(os.path.join(work_dir, logfile))

        if os.path.exists(os.path.join(work_dir, trajectory)):
            os.remove(os.path.join(work_dir, trajectory))

        md = MolecularDynamics(
            atoms,
            ensemble=ensemble,
            temperature=temperature,
            timestep=timestep,
            loginterval=loginterval,
            logfile=os.path.join(work_dir, logfile),
            trajectory=os.path.join(work_dir, trajectory),
            taut=taut,
        )
        md.run(steps)

        # parse the logfile

        # Read the file into a pandas DataFrame
        df = pd.read_csv(
            os.path.join(work_dir, logfile),
            sep="\\s+",
        )
        df.columns = list(
            map(lambda x: re.sub(r"\[.*?\]", "", x).strip().lower(), df.columns)
        )
        traj = read(os.path.join(work_dir, trajectory), index=":")
        print(df.shape)
        print(len(traj))
        structure_list = [AseAtomsAdaptor.get_structure(atoms) for atoms in traj]

        # Add the structure list to the DataFrame
        df["structure"] = [structure.to_json() for structure in structure_list]

        # Print the DataFrame
        print(df)

        # Save the DataFrame to a CSV file
        df.to_csv(os.path.join(work_dir, save_csv))

        moldyn_results = df

    return moldyn_results



# -*- coding: utf-8 -*-
import argparse
import os
import random

from ase.io import write

from mattersim.utils.atoms_utils import AtomsAdaptor


def main(args):
    vasp_files = []
    for root, dirs, files in os.walk(args.data_path):
        for file in files:
            if file.endswith(".xml"):
                vasp_files.append(os.path.join(root, file))

    train_ratio = args.train_ratio
    validation_ratio = args.validation_ratio

    save_dir = args.save_path
    os.makedirs(save_dir, exist_ok=True)

    atoms_train = []
    atoms_validation = []
    atoms_test = []

    random.seed(args.seed)

    for vasp_file in vasp_files:
        atoms_list = AtomsAdaptor.from_file(filename=vasp_file)
        random.shuffle(atoms_list)
        num_atoms = len(atoms_list)
        num_train = int(num_atoms * train_ratio)
        num_validation = int(num_atoms * validation_ratio)

        atoms_train.extend(atoms_list[:num_train])
        atoms_validation.extend(atoms_list[num_train : num_train + num_validation])
        atoms_test.extend(atoms_list[num_train + num_validation :])

    print(
        f"Total number of atoms: {len(atoms_train) + len(atoms_validation) + len(atoms_test)}"  # noqa: E501
    )

    print(f"Number of atoms in the training set: {len(atoms_train)}")
    print(f"Number of atoms in the validation set: {len(atoms_validation)}")
    print(f"Number of atoms in the test set: {len(atoms_test)}")

    # Save the training, validation, and test datasets to xyz files

    write(f"{save_dir}/train.xyz", atoms_train, format="extxyz")
    write(f"{save_dir}/valid.xyz", atoms_validation, format="extxyz")
    write(f"{save_dir}/test.xyz", atoms_test, format="extxyz")


if __name__ == "__main__":
    # Some important arguments
    parser = argparse.ArgumentParser()

    # path parameters
    parser.add_argument("--data_path", type=str, default=None, help="vasprun data path")
    parser.add_argument("--train_ratio", type=float, default=0.8, help="train ratio")
    parser.add_argument(
        "--validation_ratio", type=float, default=0.1, help="validation ratio"
    )
    parser.add_argument(
        "--save_path",
        type=str,
        default="./xyz_files",
        help="path to save the xyz files",
    )
    parser.add_argument(
        "--seed",
        type=int,
        default=42,
    )
    args = parser.parse_args()
    main(args)



# -*- coding: utf-8 -*-
import argparse
import os
import pickle as pkl
import random

import numpy as np
import torch
import torch.distributed
import wandb
from ase.units import GPa

from mattersim.datasets.utils.build import build_dataloader
from mattersim.forcefield.m3gnet.scaling import AtomScaling
from mattersim.forcefield.potential import Potential
from mattersim.utils.atoms_utils import AtomsAdaptor
from mattersim.utils.logger_utils import get_logger

logger = get_logger()
local_rank = int(os.environ["LOCAL_RANK"])


def main(args):
    if args.device == "cuda":
        torch.distributed.init_process_group(backend="nccl")
    else:
        torch.distributed.init_process_group(backend="gloo")
    args_dict = vars(args)
    if args.wandb and local_rank == 0:
        wandb_api_key = (
            args.wandb_api_key
            if args.wandb_api_key is not None
            else os.getenv("WANDB_API_KEY")
        )
        wandb.login(key=wandb_api_key)
        wandb.init(
            project=args.wandb_project,
            name=args.run_name,
            config=args,
            # id=args.run_name,
            # resume="allow",
        )

    if args.wandb:
        args_dict["wandb"] = wandb

    torch.distributed.barrier()

    # set random seed
    random.seed(args.seed)
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)

    if args.device == "cuda":
        torch.cuda.set_device(local_rank)

    if args.train_data_path.endswith(".pkl"):
        with open(args.train_data_path, "rb") as f:
            atoms_train = pkl.load(f)
    else:
        atoms_train = AtomsAdaptor.from_file(filename=args.train_data_path)
    energies = []
    forces = [] if args.include_forces else None
    stresses = [] if args.include_stresses else None
    logger.info("Processing training datasets...")
    for atoms in atoms_train:
        energies.append(atoms.get_potential_energy())
        if args.include_forces:
            forces.append(atoms.get_forces())
        if args.include_stresses:
            stresses.append(atoms.get_stress(voigt=False) / GPa)  # convert to GPa

    dataloader = build_dataloader(
        atoms_train,
        energies,
        forces,
        stresses,
        shuffle=True,
        pin_memory=(args.device == "cuda"),
        is_distributed=True,
        **args_dict,
    )

    device = args.device
    # build energy normalization module
    if args.re_normalize:
        scale = AtomScaling(
            atoms=atoms_train,
            total_energy=energies,
            forces=forces,
            verbose=True,
            **args_dict,
        ).to(device)

    if args.valid_data_path is not None:
        if args.valid_data_path.endswith(".pkl"):
            with open(args.valid_data_path, "rb") as f:
                atoms_val = pkl.load(f)
        else:
            atoms_val = AtomsAdaptor.from_file(filename=args.valid_data_path)
        energies = []
        forces = [] if args.include_forces else None
        stresses = [] if args.include_stresses else None
        logger.info("Processing validation datasets...")
        for atoms in atoms_val:
            energies.append(atoms.get_potential_energy())
            if args.include_forces:
                forces.append(atoms.get_forces())
            if args.include_stresses:
                stresses.append(atoms.get_stress(voigt=False) / GPa)  # convert to GPa
        val_dataloader = build_dataloader(
            atoms_val,
            energies,
            forces,
            stresses,
            pin_memory=(args.device == "cuda"),
            is_distributed=True,
            **args_dict,
        )
    else:
        val_dataloader = None

    potential = Potential.from_checkpoint(
        load_path=args.load_model_path,
        load_training_state=False,
        **args_dict,
    )

    if args.re_normalize:
        potential.model.set_normalizer(scale)

    if args.device == "cuda":
        potential.model = torch.nn.parallel.DistributedDataParallel(potential.model)
    torch.distributed.barrier()

    potential.train_model(
        dataloader,
        val_dataloader,
        loss=torch.nn.HuberLoss(delta=0.01),
        is_distributed=True,
        **args_dict,
    )

    if local_rank == 0 and args.save_checkpoint and args.wandb:
        wandb.save(os.path.join(args.save_path, "best_model.pth"))


if __name__ == "__main__":
    # Some important arguments
    parser = argparse.ArgumentParser()

    # path parameters
    parser.add_argument(
        "--run_name", type=str, default="example", help="name of the run"
    )
    parser.add_argument(
        "--train_data_path", type=str, default="./sample.xyz", help="train data path"
    )
    parser.add_argument(
        "--valid_data_path", type=str, default=None, help="valid data path"
    )
    parser.add_argument(
        "--load_model_path",
        type=str,
        default="mattersim-v1.0.0-1m",
        help="path to load the model",
    )
    parser.add_argument(
        "--save_path", type=str, default="./results", help="path to save the model"
    )
    parser.add_argument(
        "--save_checkpoint",
        type=bool,
        default=False,
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        "--ckpt_interval",
        type=int,
        default=10,
        help="save checkpoint every ckpt_interval epochs",
    )
    parser.add_argument("--device", type=str, default="cuda")

    # model parameters
    parser.add_argument("--cutoff", type=float, default=5.0, help="cutoff radius")
    parser.add_argument(
        "--threebody_cutoff",
        type=float,
        default=4.0,
        help="cutoff radius for three-body term, which should be smaller than cutoff (two-body)",  # noqa: E501
    )

    # training parameters
    parser.add_argument("--epochs", type=int, default=1000, help="number of epochs")
    parser.add_argument("--batch_size", type=int, default=16)
    parser.add_argument("--lr", type=float, default=2e-4)
    parser.add_argument(
        "--step_size",
        type=int,
        default=10,
        help="step epoch for learning rate scheduler",
    )
    parser.add_argument(
        "--include_forces",
        type=bool,
        default=True,
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        "--include_stresses",
        type=bool,
        default=False,
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument("--force_loss_ratio", type=float, default=1.0)
    parser.add_argument("--stress_loss_ratio", type=float, default=0.1)
    parser.add_argument("--early_stop_patience", type=int, default=10)
    parser.add_argument("--seed", type=int, default=42)

    # scaling parameters
    parser.add_argument(
        "--re_normalize",
        type=bool,
        default=False,
        action=argparse.BooleanOptionalAction,
        help="re-normalize the energy and forces according to the new data",
    )
    parser.add_argument("--scale_key", type=str, default="per_species_forces_rms")
    parser.add_argument(
        "--shift_key", type=str, default="per_species_energy_mean_linear_reg"
    )
    parser.add_argument("--init_scale", type=float, default=None)
    parser.add_argument("--init_shift", type=float, default=None)
    parser.add_argument(
        "--trainable_scale",
        type=bool,
        default=False,
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        "--trainable_shift",
        type=bool,
        default=False,
        action=argparse.BooleanOptionalAction,
    )

    # wandb parameters
    parser.add_argument("--wandb", action="store_true")
    parser.add_argument("--wandb_api_key", type=str, default=None)
    parser.add_argument("--wandb_project", type=str, default="wandb_test")
    args = parser.parse_args()
    main(args)



# -*- coding: utf-8 -*-
from typing import Dict

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_runstats.scatter import scatter

from mattersim.jit_compile_tools.jit import compile_mode

from .modules import (  # noqa: F501
    MLP,
    GatedMLP,
    MainBlock,
    SmoothBesselBasis,
    SphericalBasisLayer,
)
from .scaling import AtomScaling


@compile_mode("script")
class M3Gnet(nn.Module):
    """
    M3Gnet
    """

    def __init__(
        self,
        num_blocks: int = 4,
        units: int = 128,
        max_l: int = 4,
        max_n: int = 4,
        cutoff: float = 5.0,
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
        max_z: int = 94,
        threebody_cutoff: float = 4.0,
        **kwargs,
    ):
        super().__init__()
        self.rbf = SmoothBesselBasis(r_max=cutoff, max_n=max_n)
        self.sbf = SphericalBasisLayer(max_n=max_n, max_l=max_l, cutoff=cutoff)
        self.edge_encoder = MLP(
            in_dim=max_n, out_dims=[units], activation="swish", use_bias=False
        )
        module_list = [
            MainBlock(max_n, max_l, cutoff, units, max_n, threebody_cutoff)
            for i in range(num_blocks)
        ]
        self.graph_conv = nn.ModuleList(module_list)
        self.final = GatedMLP(
            in_dim=units,
            out_dims=[units, units, 1],
            activation=["swish", "swish", None],
        )
        self.apply(self.init_weights)
        self.atom_embedding = MLP(
            in_dim=max_z + 1, out_dims=[units], activation=None, use_bias=False
        )
        self.atom_embedding.apply(self.init_weights_uniform)
        self.normalizer = AtomScaling(verbose=False, max_z=max_z)
        self.max_z = max_z
        self.device = device
        self.model_args = {
            "num_blocks": num_blocks,
            "units": units,
            "max_l": max_l,
            "max_n": max_n,
            "cutoff": cutoff,
            "max_z": max_z,
            "threebody_cutoff": threebody_cutoff,
        }

    def forward(
        self,
        input: Dict[str, torch.Tensor],
        dataset_idx: int = -1,
    ) -> torch.Tensor:
        # Exact data from input_dictionary
        pos = input["atom_pos"]
        cell = input["cell"]
        pbc_offsets = input["pbc_offsets"].float()
        atom_attr = input["atom_attr"]
        edge_index = input["edge_index"].long()
        three_body_indices = input["three_body_indices"].long()
        num_three_body = input["num_three_body"]
        num_bonds = input["num_bonds"]
        num_triple_ij = input["num_triple_ij"]
        num_atoms = input["num_atoms"]
        num_graphs = input["num_graphs"]
        batch = input["batch"]

        # -------------------------------------------------------------#
        cumsum = torch.cumsum(num_bonds, dim=0) - num_bonds
        index_bias = torch.repeat_interleave(  # noqa: F501
            cumsum, num_three_body, dim=0
        ).unsqueeze(-1)
        three_body_indices = three_body_indices + index_bias

        # === Refer to the implementation of M3GNet,        ===
        # === we should re-compute the following attributes ===
        # edge_length, edge_vector(optional), triple_edge_length, theta_jik
        atoms_batch = torch.repeat_interleave(repeats=num_atoms)
        edge_batch = atoms_batch[edge_index[0]]
        edge_vector = pos[edge_index[0]] - (
            pos[edge_index[1]]
            + torch.einsum("bi, bij->bj", pbc_offsets, cell[edge_batch])
        )
        edge_length = torch.linalg.norm(edge_vector, dim=1)
        vij = edge_vector[three_body_indices[:, 0].clone()]
        vik = edge_vector[three_body_indices[:, 1].clone()]
        rij = edge_length[three_body_indices[:, 0].clone()]
        rik = edge_length[three_body_indices[:, 1].clone()]
        cos_jik = torch.sum(vij * vik, dim=1) / (rij * rik)
        # eps = 1e-7 avoid nan in torch.acos function
        cos_jik = torch.clamp(cos_jik, min=-1.0 + 1e-7, max=1.0 - 1e-7)
        triple_edge_length = rik.view(-1)
        edge_length = edge_length.unsqueeze(-1)
        atomic_numbers = atom_attr.squeeze(1).long()

        # featurize
        atom_attr = self.atom_embedding(self.one_hot_atoms(atomic_numbers))
        edge_attr = self.rbf(edge_length.view(-1))
        edge_attr_zero = edge_attr  # e_ij^0
        edge_attr = self.edge_encoder(edge_attr)
        three_basis = self.sbf(triple_edge_length, torch.acos(cos_jik))

        # Main Loop
        for idx, conv in enumerate(self.graph_conv):
            atom_attr, edge_attr = conv(
                atom_attr,
                edge_attr,
                edge_attr_zero,
                edge_index,
                three_basis,
                three_body_indices,
                edge_length,
                num_bonds,
                num_triple_ij,
                num_atoms,
            )

        energies_i = self.final(atom_attr).view(-1)  # [batch_size*num_atoms]
        energies_i = self.normalizer(energies_i, atomic_numbers)
        energies = scatter(energies_i, batch, dim=0, dim_size=num_graphs)

        return energies  # [batch_size]

    def init_weights(self, m):
        if isinstance(m, nn.Linear):
            torch.nn.init.xavier_uniform_(m.weight)

    def init_weights_uniform(self, m):
        if isinstance(m, nn.Linear):
            torch.nn.init.uniform_(m.weight, a=-0.05, b=0.05)

    @torch.jit.export
    def one_hot_atoms(self, species):
        # one_hots = []
        # for i in range(species.shape[0]):
        #     one_hots.append(
        #         F.one_hot(
        #             species[i],
        #             num_classes=self.max_z+1).float().to(species.device)
        #     )
        # return torch.cat(one_hots, dim=0)
        return F.one_hot(species, num_classes=self.max_z + 1).float()

    def print(self):
        from prettytable import PrettyTable

        table = PrettyTable(["Modules", "Parameters"])
        total_params = 0
        for name, parameter in self.named_parameters():
            if not parameter.requires_grad:
                continue
            params = parameter.numel()
            table.add_row([name, params])
            total_params += params
        print(table)
        print(f"Total Trainable Params: {total_params}")

    @torch.jit.export
    def set_normalizer(self, normalizer: AtomScaling):
        self.normalizer = normalizer

    def get_model_args(self):
        return self.model_args



# -*- coding: utf-8 -*-
"""
Ref:
    - https://github.com/mir-group/nequip
    - https://www.nature.com/articles/s41467-022-29939-5
"""

import math
from typing import Optional

import torch
from e3nn.math import soft_one_hot_linspace
from torch import nn

from mattersim.jit_compile_tools.jit import compile_mode


class e3nn_basias(nn.Module):
    def __init__(
        self,
        r_max: float,
        r_min: Optional[float] = None,
        e3nn_basis_name: str = "gaussian",
        num_basis: int = 8,
    ):
        super().__init__()
        self.r_max = r_max
        self.r_min = r_min if r_min is not None else 0.0
        self.e3nn_basis_name = e3nn_basis_name
        self.num_basis = num_basis

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return soft_one_hot_linspace(
            x,
            start=self.r_min,
            end=self.r_max,
            number=self.num_basis,
            basis=self.e3nn_basis_name,
            cutoff=True,
        )

    def _make_tracing_inputs(self, n: int):
        return [{"forward": (torch.randn(5, 1),)} for _ in range(n)]


class BesselBasis(nn.Module):
    def __init__(self, r_max, num_basis=8, trainable=True):
        r"""Radial Bessel Basis, as proposed in
            DimeNet: https://arxiv.org/abs/2003.03123

        Parameters
        ----------
        r_max : float
            Cutoff radius

        num_basis : int
            Number of Bessel Basis functions

        trainable : bool
            Train the :math:`n \pi` part or not.
        """
        super(BesselBasis, self).__init__()

        self.trainable = trainable
        self.num_basis = num_basis

        self.r_max = float(r_max)
        self.prefactor = 2.0 / self.r_max

        bessel_weights = (
            torch.linspace(start=1.0, end=num_basis, steps=num_basis) * math.pi
        )
        if self.trainable:
            self.bessel_weights = nn.Parameter(bessel_weights)
        else:
            self.register_buffer("bessel_weights", bessel_weights)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Evaluate Bessel Basis for input x.

        Parameters
        ----------
        x : torch.Tensor
            Input
        """
        numerator = torch.sin(
            self.bessel_weights * x.unsqueeze(-1) / self.r_max  # noqa: E501
        )

        return self.prefactor * (numerator / x.unsqueeze(-1))


@compile_mode("script")
class SmoothBesselBasis(nn.Module):
    def __init__(self, r_max, max_n=10):
        r"""Smooth Radial Bessel Basis, as proposed
            in DimeNet: https://arxiv.org/abs/2003.03123
            This is an orthogonal basis with first
            and second derivative at the cutoff
            equals to zero. The function was derived from
            the order 0 spherical Bessel function,
            and was expanded by the different zero roots
        Ref:
            https://arxiv.org/pdf/1907.02374.pdf
        Args:
            r_max: torch.Tensor distance tensor
            max_n: int, max number of basis, expanded by the zero roots
        Returns: expanded spherical harmonics with
                 derivatives smooth at boundary
        Parameters
        ----------
        """
        super(SmoothBesselBasis, self).__init__()
        self.max_n = max_n
        n = torch.arange(0, max_n).float()[None, :]
        PI = 3.1415926535897
        SQRT2 = 1.41421356237
        fnr = (
            (-1) ** n
            * SQRT2
            * PI
            / r_max**1.5
            * (n + 1)
            * (n + 2)
            / torch.sqrt(2 * n**2 + 6 * n + 5)
        )
        en = n**2 * (n + 2) ** 2 / (4 * (n + 1) ** 4 + 1)
        dn = [torch.tensor(1.0).float()]
        for i in range(1, max_n):
            dn.append(1 - en[0, i] / dn[-1])
        dn = torch.stack(dn)
        self.register_buffer("dn", dn)
        self.register_buffer("en", en)
        self.register_buffer("fnr_weights", fnr)
        self.register_buffer(
            "n_1_pi_cutoff",
            ((torch.arange(0, max_n).float() + 1) * PI / r_max).reshape(1, -1),
        )
        self.register_buffer(
            "n_2_pi_cutoff",
            ((torch.arange(0, max_n).float() + 2) * PI / r_max).reshape(1, -1),
        )
        self.register_buffer("r_max", torch.tensor(r_max))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Evaluate Smooth Bessel Basis for input x.

        Parameters
        ----------
        x : torch.Tensor
            Input
        """
        x_1 = x.unsqueeze(-1) * self.n_1_pi_cutoff
        x_2 = x.unsqueeze(-1) * self.n_2_pi_cutoff
        fnr = self.fnr_weights * (torch.sin(x_1) / x_1 + torch.sin(x_2) / x_2)
        gn = [fnr[:, 0]]
        for i in range(1, self.max_n):
            gn.append(
                1
                / torch.sqrt(self.dn[i])
                * (
                    fnr[:, i]
                    + torch.sqrt(self.en[0, i] / self.dn[i - 1]) * gn[-1]  # noqa: E501
                )
            )
        return torch.transpose(torch.stack(gn), 1, 0)


# class GaussianBasis(nn.Module):
#     r_max: float

#     def __init__(self, r_max, r_min=0.0, num_basis=8, trainable=True):
#         super().__init__()

#         self.trainable = trainable
#         self.num_basis = num_basis

#         self.r_max = float(r_max)
#         self.r_min = float(r_min)

#         means = torch.linsspace(self.r_min, self.r_max, self.num_basis)
#         stds = torch.full(size=means.size, fill_value=means[1] - means[0])
#         if self.trainable:
#             self.means = nn.Parameter(means)
#             self.stds = nn.Parameter(stds)
#         else:
#             self.register_buffer("means", means)
#             self.register_buffer("stds", stds)

#     def forward(self, x: torch.Tensor) -> torch.Tensor:
#         x = (x[..., None] - self.means) / self.stds
#         x = x.square().mul(-0.5).exp() / self.stds  # sqrt(2 * pi)



# -*- coding: utf-8 -*-
import warnings
from typing import Iterable, List, Tuple, Union

from ase import Atoms
from ase.constraints import Filter, FixSymmetry
from ase.filters import ExpCellFilter, FrechetCellFilter
from ase.optimize import BFGS, FIRE
from ase.optimize.optimize import Optimizer
from ase.units import GPa
from deprecated import deprecated


class Relaxer(object):
    """Relaxer is a class for structural relaxation with fixed volume."""

    SUPPORTED_OPTIMIZERS = {"BFGS": BFGS, "FIRE": FIRE}
    SUPPORTED_FILTERS = {
        "EXPCELLFILTER": ExpCellFilter,
        "FRECHETCELLFILTER": FrechetCellFilter,
    }

    def __init__(
        self,
        optimizer: Union[Optimizer, str] = "FIRE",
        filter: Union[Filter, str, None] = None,
        constrain_symmetry: bool = True,
        fix_axis: Union[bool, Iterable[bool]] = False,
    ) -> None:
        """
        Args:
            optimizer (Union[Optimizer, str]): The optimizer to use.
            filter (Union[Filter, str, None]): The filter to use.
            constrain_symmetry (bool): Whether to constrain the symmetry.
            fix_axis (Union[bool, Iterable[bool]]): Whether to fix the axis.
        """
        self.optimizer = (
            self.SUPPORTED_OPTIMIZERS[optimizer.upper()]
            if isinstance(optimizer, str)
            else optimizer
        )
        self.relax_cell = filter is not None
        if filter is not None:
            self.filter = (
                self.SUPPORTED_FILTERS[filter.upper()]
                if isinstance(filter, str)
                else filter
            )
        self.constrain_symmetry = constrain_symmetry
        self.fix_axis = fix_axis

    def relax(
        self,
        atoms: Atoms,
        steps: int = 500,
        fmax: float = 0.01,
        params_filter: dict = {},
        **kwargs,
    ) -> Tuple[bool, Atoms]:
        """
        Relax the atoms object.

        Args:
            atoms (Atoms): The atoms object to relax.
            steps (int): The maximum number of steps to take.
            fmax (float): The maximum force allowed.
            params_filter (dict): The parameters for the filter.
            kwargs: Additional keyword arguments for the optimizer.
        """

        if atoms.calc is None:
            raise ValueError("Atoms object must have a calculator.")

        if self.constrain_symmetry:
            atoms.set_constraint(FixSymmetry(atoms))

        if self.relax_cell:
            # Set the mask for the fixed axis
            if isinstance(self.fix_axis, bool):
                mask = [not self.fix_axis for i in range(6)]
            else:
                assert (
                    len(self.fix_axis) == 6
                ), "The length of fix_axis list not equal 6."
                mask = [not elem for elem in self.fix_axis]

            # check if the scalar_pressure is provided
            if (
                "scalar_pressure" in params_filter
                and params_filter["scalar_pressure"] > 1
            ):
                warnings.warn(
                    "The scalar_pressure used in ExpCellFilter assumes "
                    "eV/A^3 unit and 1 eV/A^3 is already 160 GPa. "
                    "Please make sure you have converted your pressure "
                    "from GPa to eV/A^3 by dividing by 160.21766208."
                )
            ecf = self.filter(atoms, mask=mask, **params_filter)
        else:
            ecf = atoms
        optimizer = self.optimizer(ecf, **kwargs)
        optimizer.run(fmax=fmax, steps=steps)

        converged = optimizer.get_number_of_steps() < steps

        if self.constrain_symmetry:
            atoms.set_constraint(None)

        return converged, atoms

    @classmethod
    @deprecated(reason="Use cli/applications/relax_structure.py instead.")
    def relax_structures(
        cls,
        atoms: Union[Atoms, Iterable[Atoms]],
        optimizer: Union[Optimizer, str] = "FIRE",
        filter: Union[Filter, str, None] = None,
        constrain_symmetry: bool = False,
        fix_axis: Union[bool, Iterable[bool]] = False,
        pressure_in_GPa: Union[float, None] = None,
        **kwargs,
    ) -> Union[Tuple[bool, Atoms], Tuple[List[bool], List[Atoms]]]:
        """
        Args:
            atoms: (Union[Atoms, Iterable[Atoms]]):
                The Atoms object or an iterable of Atoms objects to relax.
            optimizer (Union[Optimizer, str]): The optimizer to use.
            filter (Union[Filter, str, None]): The filter to use.
            constrain_symmetry (bool): Whether to constrain the symmetry.
            fix_axis (Union[bool, Iterable[bool]]): Whether to fix the axis.
            **kwargs: Additional keyword arguments for the relax method.
        Returns:
            converged (Union[bool, List[bool]]):
                Whether the relaxation converged or a list of them
            Atoms (Union[Atoms, List[Atoms]]):
                The relaxed atoms object or a list of them
        """
        params_filter = {}

        if filter is None and pressure_in_GPa is None:
            pass
        elif filter is None and pressure_in_GPa is not None:
            filter = "ExpCellFilter"
            params_filter["scalar_pressure"] = (
                pressure_in_GPa * GPa
            )  # GPa = 1 / 160.21766208
        elif filter is not None and pressure_in_GPa is None:
            params_filter["scalar_pressure"] = 0.0
        else:
            params_filter["scalar_pressure"] = (
                pressure_in_GPa * GPa
            )  # GPa = / 160.21766208

        relaxer = Relaxer(
            optimizer=optimizer,
            filter=filter,
            constrain_symmetry=constrain_symmetry,
            fix_axis=fix_axis,
        )

        if isinstance(atoms, (list, tuple)):
            relaxed_results = relaxed_results = [
                relaxer.relax(atom, params_filter=params_filter, **kwargs)
                for atom in atoms
            ]
            converged, relaxed_atoms = zip(*relaxed_results)
            return list(converged), list(relaxed_atoms)
        else:
            return relaxer.relax(atoms, params_filter=params_filter, **kwargs)



# -*- coding: utf-8 -*-
import unittest

import numpy as np
from ase import Atoms

from mattersim.applications.batch_relax import BatchRelaxer
from mattersim.forcefield.potential import Potential


class RelaxerTestCase(unittest.TestCase):
    def setUp(self):
        # Create an example structure for testing
        a = 3.567
        positions = [
            (0, 0, 0),
            (a / 4, a / 4, a / 4),
            (a / 2, a / 2, 0),
            (a / 2, 0, a / 2),
            (0, a / 2, a / 2),
            (a / 4, 3 * a / 4, 3 * a / 4),
            (3 * a / 4, a / 4, 3 * a / 4),
            (3 * a / 4, 3 * a / 4, a / 4),
        ]
        cell = [(a, 0, 0), (0, a, 0), (0, 0, a)]
        self.atoms_ideal = Atoms(
            "C8", positions=positions, cell=cell, pbc=True  # noqa: E501
        )

        # Create an example structure with displaced atoms for testing
        a = 3.567  # Angstroms
        positions = [
            (0, 0, 0),
            (a / 4, a / 4, a / 4),
            (a / 2, a / 2, 0),
            (a / 2, 0, a / 2),
            (0, a / 2, a / 2),
            (a / 4, 3 * a / 4, 3 * a / 4.01),  # displaced
            (3 * a / 4, a / 4.01, 3 * a / 4),  # displaced
            (3 * a / 4, 3 * a / 4, a / 4),
        ]
        cell = [(a, 0, 0), (0, a, 0), (0, 0, a)]
        self.atoms_displaced = Atoms(
            "C8", positions=positions, cell=cell, pbc=True  # noqa: E501
        )

        # Create an example structure with expanded cell for testing
        a = 3.567 * 1.2
        positions = [
            (0, 0, 0),
            (a / 4, a / 4, a / 4),
            (a / 2, a / 2, 0),
            (a / 2, 0, a / 2),
            (0, a / 2, a / 2),
            (a / 4, 3 * a / 4, 3 * a / 4),
            (3 * a / 4, a / 4, 3 * a / 4),
            (3 * a / 4, 3 * a / 4, a / 4),
        ]
        cell = [(a, 0, 0), (0, a, 0), (0, 0, a)]
        self.atoms_expanded = Atoms(
            "C8", positions=positions, cell=cell, pbc=True  # noqa: E501
        )
        # Create a batch of structures for testing
        self.atoms_batch = [self.atoms_ideal, self.atoms_displaced, self.atoms_expanded]

        self.potential = Potential.from_checkpoint()

    def test_default_batch_relaxer(self):
        relaxer = BatchRelaxer(self.potential, fmax=0.01, filter="EXPCELLFILTER")
        atoms_batch = self.atoms_batch.copy()
        relaxation_trajectories = relaxer.relax(atoms_batch)
        assert len(relaxation_trajectories) == len(atoms_batch)
        relaxed_ideal = relaxation_trajectories[0][-1]
        for trajectory in relaxation_trajectories.values():
            assert len(trajectory) > 0
            assert trajectory[-1].info["total_energy"] is not None
            assert trajectory[-1].arrays["forces"] is not None
            assert trajectory[-1].info["stress"] is not None
            assert np.allclose(trajectory[-1].get_positions(), relaxed_ideal.get_positions(), atol=0.01)
            assert np.allclose(trajectory[-1].get_cell(), relaxed_ideal.get_cell(), atol=0.01)


if __name__ == "__main__":
    unittest.main()



import os
import uuid
from collections import defaultdict
from typing import List

import pandas as pd
from ase import Atoms
from ase.units import GPa
from loguru import logger
from pymatgen.io.ase import AseAtomsAdaptor
from tqdm import tqdm


def singlepoint(
    atoms_list: List[Atoms],
    *,
    work_dir: str = str(uuid.uuid4()),
    save_csv: str = "results.csv.gz",
    **kwargs,
) -> dict:
    """
    Predict single point properties for a list of atoms.

    """
    logger.info("Predicting single point properties.")
    predicted_properties = defaultdict(list)
    for atoms in tqdm(
        atoms_list, total=len(atoms_list), desc="Predicting single point properties"
    ):
        predicted_properties["structure"].append(AseAtomsAdaptor.get_structure(atoms).as_dict())
        predicted_properties["energy"].append(atoms.get_potential_energy())
        predicted_properties["energy_per_atom"].append(
            atoms.get_potential_energy() / len(atoms)
        )
        predicted_properties["forces"].append(atoms.get_forces())
        predicted_properties["stress"].append(atoms.get_stress(voigt=False))
        predicted_properties["stress_GPa"].append(atoms.get_stress(voigt=False) / GPa)

    if not os.path.exists(work_dir):
        os.makedirs(work_dir)

    logger.info(f"Saving the results to {os.path.join(work_dir, save_csv)}")

    df = pd.DataFrame(predicted_properties)
    df.to_csv(os.path.join(work_dir, save_csv), index=False, mode="a")
    return predicted_properties



# -*- coding: utf-8 -*-
import torch
import torch.nn as nn
from torch_runstats.scatter import scatter

from .layers import GatedMLP, LinearLayer, SigmoidLayer, SwishLayer


def polynomial(r: torch.Tensor, cutoff: float) -> torch.Tensor:
    """
    Polynomial cutoff function
    Args:
        r (tf.Tensor): radius distance tensor
        cutoff (float): cutoff distance
    Returns: polynomial cutoff functions
    """
    ratio = torch.div(r, cutoff)
    result = (
        1
        - 6 * torch.pow(ratio, 5)
        + 15 * torch.pow(ratio, 4)
        - 10 * torch.pow(ratio, 3)
    )
    return torch.clamp(result, min=0.0)


class ThreeDInteraction(nn.Module):
    def __init__(
        self,
        max_n,
        max_l,
        cutoff,
        units,
        spherecal_dim,
        threebody_cutoff,
    ):
        super().__init__()
        # self.sbf = SphericalBesselFunction(
        #            max_l=max_l, max_n=max_n, cutoff=cutoff, smooth=smooth)
        # self.shf = SphericalHarmonicsFunction(max_l=max_l, use_phi=use_phi)
        self.atom_mlp = SigmoidLayer(in_dim=units, out_dim=spherecal_dim)
        # Linyu have modified the self.edge_gate_mlp
        # by adding swish activation and use_bias=False
        self.edge_gate_mlp = GatedMLP(
            in_dim=spherecal_dim,
            out_dims=[units],
            activation="swish",
            use_bias=False,  # noqa: E501
        )
        self.cutoff = cutoff
        self.threebody_cutoff = threebody_cutoff

    def forward(
        self,
        edge_attr,
        three_basis,
        atom_attr,
        edge_index,
        three_body_index,
        edge_length,
        num_edges,
        num_triple_ij,
    ):
        atom_mask = (
            self.atom_mlp(atom_attr)[edge_index[0][three_body_index[:, 1]]]
            * polynomial(
                edge_length[three_body_index[:, 0]], self.threebody_cutoff  # noqa: E501
            )
            * polynomial(
                edge_length[three_body_index[:, 1]], self.threebody_cutoff  # noqa: E501
            )
        )
        three_basis = three_basis * atom_mask
        index_map = torch.arange(torch.sum(num_edges).item()).to(
            edge_length.device
        )  # noqa: E501
        index_map = torch.repeat_interleave(index_map, num_triple_ij).to(
            edge_length.device
        )
        e_ij_tuda = scatter(
            three_basis,
            index_map,
            dim=0,
            reduce="sum",
            dim_size=torch.sum(num_edges).item(),
        )
        edge_attr_prime = edge_attr + self.edge_gate_mlp(e_ij_tuda)
        return edge_attr_prime


class AtomLayer(nn.Module):
    """
    v_i'=v_i+sum(phi(v+i,v_j,e_ij',u)W*e_ij^0)
    """

    def __init__(
        self,
        atom_attr_dim,
        edge_attr_dim,
        spherecal_dim,
    ):
        super().__init__()
        self.gated_mlp = GatedMLP(
            in_dim=2 * atom_attr_dim + spherecal_dim,
            out_dims=[128, 64, atom_attr_dim],  # noqa: E501
        )  # [2*atom_attr_dim+edge_attr_prime_dim]  ->  [atom_attr_dim]
        self.edge_layer = LinearLayer(
            in_dim=edge_attr_dim, out_dim=1
        )  # [atom_attr_dim]  ->  [1]

    def forward(
        self,
        atom_attr,
        edge_attr,
        edge_index,
        edge_attr_prime,  # [sum(num_edges),edge_attr_dim]
        num_atoms,  # [batch_size]
    ):
        feat = torch.concat(
            [
                atom_attr[edge_index[0]],
                atom_attr[edge_index[1]],
                edge_attr_prime,
            ],  # noqa: E501
            dim=1,
        )
        atom_attr_prime = self.gated_mlp(feat) * self.edge_layer(edge_attr)
        atom_attr_prime = scatter(
            atom_attr_prime,
            edge_index[1],
            dim=0,
            dim_size=torch.sum(num_atoms).item(),  # noqa: E501
        )
        return atom_attr_prime + atom_attr


class EdgeLayer(nn.Module):
    """e_ij'=e_ij+phi(v_i,v_j,e_ij,u)W*e_ij^0"""

    def init(
        self,
        atom_attr_dim,
        edge_attr_dim,
        spherecal_dim,
    ):
        super().__init__()
        self.gated_mlp = GatedMLP(
            in_dim=2 * atom_attr_dim + spherecal_dim,
            out_dims=[128, 64, edge_attr_dim],  # noqa: E501
        )
        self.edge_layer = LinearLayer(in_dim=edge_attr_dim, out_dim=1)

    def forward(
        self,
        atom_attr,
        edge_attr,
        edge_index,
        edge_attr_prime,  # [sum(num_edges),edge_attr_dim]
    ):
        feat = torch.concat(
            [
                atom_attr[edge_index[0]],
                atom_attr[edge_index[1]],
                edge_attr_prime,
            ],  # noqa: E501
            dim=1,
        )
        edge_attr_prime = self.gated_mlp(feat) * self.edge_layer(edge_attr)
        return edge_attr_prime + edge_attr


class MainBlock(nn.Module):
    """
    MainBlock for Message Passing in M3GNet
    """

    def __init__(
        self,
        max_n,
        max_l,
        cutoff,
        units,
        spherical_dim,
        threebody_cutoff,
    ):
        super().__init__()
        self.gated_mlp_atom = GatedMLP(
            in_dim=2 * units + units,
            out_dims=[units, units],
            activation="swish",  # noqa: E501
        )  # [2*atom_attr_dim+edge_attr_prime_dim]  ->  [units]
        self.edge_layer_atom = SwishLayer(
            in_dim=spherical_dim, out_dim=units, bias=False  # noqa: E501
        )  # [spherecal_dim]  ->  [units]
        self.gated_mlp_edge = GatedMLP(
            in_dim=2 * units + units,
            out_dims=[units, units],
            activation="swish",  # noqa: E501
        )  # [2*atom_attr_dim+edge_attr_prime_dim]  ->  [units]
        self.edge_layer_edge = LinearLayer(
            in_dim=spherical_dim, out_dim=units, bias=False
        )  # [spherecal_dim]  ->  [units]
        self.three_body = ThreeDInteraction(
            max_n, max_l, cutoff, units, max_n * max_l, threebody_cutoff
        )

    def forward(
        self,
        atom_attr,
        edge_attr,
        edge_attr_zero,
        edge_index,
        three_basis,
        three_body_index,
        edge_length,
        num_edges,
        num_triple_ij,
        num_atoms,
    ):
        # threebody interaction
        edge_attr = self.three_body(
            edge_attr,
            three_basis,
            atom_attr,
            edge_index,
            three_body_index,
            edge_length,
            num_edges,
            num_triple_ij.view(-1),
        )

        # update bond feature
        feat = torch.concat(
            [atom_attr[edge_index[0]], atom_attr[edge_index[1]], edge_attr],
            dim=1,  # noqa: E501
        )
        edge_attr = edge_attr + self.gated_mlp_edge(
            feat
        ) * self.edge_layer_edge(  # noqa: E501
            edge_attr_zero
        )

        # update atom feature
        feat = torch.concat(
            [atom_attr[edge_index[0]], atom_attr[edge_index[1]], edge_attr],
            dim=1,  # noqa: E501
        )
        atom_attr_prime = self.gated_mlp_atom(feat) * self.edge_layer_atom(
            edge_attr_zero
        )
        atom_attr = atom_attr + scatter(  # noqa: E501
            atom_attr_prime,
            edge_index[0],
            dim=0,
            dim_size=torch.sum(num_atoms).item(),  # noqa: E501
        )

        return atom_attr, edge_attr



from setuptools import setup, Extension
import numpy as np
from Cython.Build import cythonize

extensions = [
    Extension(
        "mattersim.datasets.utils.threebody_indices",
        ["src/mattersim/datasets/utils/threebody_indices.pyx"],
        include_dirs=[np.get_include()],
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_24_API_VERSION")]
    )
]

setup(
    ext_modules=cythonize(extensions)
)


# -*- coding: utf-8 -*-
import datetime
import os
from typing import Iterable, Union

import numpy as np
from ase import Atoms
from phonopy import Phonopy
from tqdm import tqdm

from mattersim.utils.phonon_utils import (
    get_primitive_cell,
    to_ase_atoms,
    to_phonopy_atoms,
)
from mattersim.utils.supercell_utils import get_supercell_parameters


class PhononWorkflow(object):
    """
    This class is used to calculate the phonon dispersion relationship of
    material using phonopy
    """

    def __init__(
        self,
        atoms: Atoms,
        find_prim: bool = False,
        work_dir: str = None,
        amplitude: float = 0.01,
        supercell_matrix: np.ndarray = None,
        qpoints_mesh: np.ndarray = None,
        max_atoms: int = None,
        calc_spec: bool = True,
    ):
        """_summary

        Args:
            atoms (Atoms): ASE atoms object contains structure information and
                calculator.
            find_prim (bool, optional): If find the primitive cell and use it
                to calculate phonon. Default to False.
            work_dir (str, optional): workplace path to contain phonon result.
                Defaults to data + chemical_symbols + 'phonon'
            amplitude (float, optional): Magnitude of the finite difference to
                displace in force constant calculation, in Angstrom. Defaults
                to 0.01 Angstrom.
            supercell_matrix (nd.array, optional): Supercell matrix for constr
                -uct supercell, priority over than max_atoms. Defaults to None.
            qpoints_mesh (nd.array, optional): Qpoint mesh for IBZ integral,
                priority over than max_atoms. Defaults to None.
            max_atoms (int, optional): Maximum atoms number limitation for the
                supercell generation. If not set, will automatic generate super
                -cell based on symmetry. Defaults to None.
            calc_spec (bool, optional): If calculate the spectrum and check 
                imaginary frequencies. Default to True.
        """
        assert (
            atoms.calc is not None
        ), "PhononWorkflow only accepts ase atoms with an attached calculator"
        if find_prim:
            self.atoms = get_primitive_cell(atoms)
            self.atoms.calc = atoms.calc
        else:
            self.atoms = atoms
        if work_dir is not None:
            self.work_dir = work_dir
        else:
            current_datetime = datetime.datetime.now()
            formatted_datetime = current_datetime.strftime("%Y-%m-%d-%H-%M")
            self.work_dir = (
                f"{formatted_datetime}-{atoms.get_chemical_formula()}-phonon"
            )
        self.amplitude = amplitude
        if supercell_matrix is not None:
            if supercell_matrix.shape == (3, 3):
                self.supercell_matrix = supercell_matrix
            elif supercell_matrix.shape == (3,):
                self.supercell_matrix = np.diag(supercell_matrix)
            else:
                assert (
                    False
                ), "Supercell matrix must be an array (3,1) or a matrix (3,3)."
        else:
            self.supercell_matrix = supercell_matrix

        if qpoints_mesh is not None:
            assert qpoints_mesh.shape == (3,), "Qpoints mesh must be an array (3,1)."
            self.qpoints_mesh = qpoints_mesh
        else:
            self.qpoints_mesh = qpoints_mesh

        self.max_atoms = max_atoms
        self.calc_spec = calc_spec

    def compute_force_constants(self, atoms: Atoms, nrep_second: np.ndarray):
        """
        Calculate force constants

        Args:
            atoms (Atoms): ASE atoms object to provide lattice informations.
            nrep_second (np.ndarray): Supercell size used for 2nd force
                constant calculations.
        """
        print(f"Supercell matrix for 2nd force constants : \n{nrep_second}")
        # Generate phonopy object
        phonon = Phonopy(
            to_phonopy_atoms(atoms),
            supercell_matrix=nrep_second,
            primitive_matrix="auto",
            log_level=2,
        )

        # Generate displacements
        phonon.generate_displacements(distance=self.amplitude)

        # Compute force constants
        second_scs = phonon.supercells_with_displacements
        second_force_sets = []
        print("\n")
        print("Inferring forces for displaced atoms and computing fcs ...")
        for disp_second in tqdm(second_scs):
            pa_second = to_ase_atoms(disp_second)
            pa_second.calc = self.atoms.calc
            second_force_sets.append(pa_second.get_forces())

        phonon.forces = np.array(second_force_sets)
        phonon.produce_force_constants()
        phonon.symmetrize_force_constants()

        return phonon

    @staticmethod
    def compute_phonon_spectrum_dos(
        atoms: Atoms, phonon: Phonopy, k_point_mesh: Union[int, Iterable[int]]
    ):
        """
        Calculate phonon spectrum and DOS based on force constant matrix in
        phonon object

        Args:
            atoms (Atoms): ASE atoms object to provide lattice information
            phonon (Phonopy): Phonopy object which contains force constants matrix
            k_point_mesh (Union[int, Iterable[int]]): The qpoints number in First
                Brillouin Zone in three directions for DOS calculation.
        """
        print(f"Qpoints mesh for Brillouin Zone integration : {k_point_mesh}")
        phonon.run_mesh(k_point_mesh)
        print(
            "Dispersion relations using phonopy for "
            + str(atoms.symbols)
            + " ..."
            + "\n"
        )

        # plot phonon spectrum
        phonon.auto_band_structure(plot=True, write_yaml=True, with_eigenvectors=True).savefig(
            f"{str(atoms.symbols)}_phonon_band.png", dpi=300
        )
        phonon.auto_total_dos(plot=True, write_dat=True).savefig(
            f"{str(atoms.symbols)}_phonon_dos.png", dpi=300
        )

        # Save additional files
        phonon.save(settings={"force_constants": True})

    @staticmethod
    def check_imaginary_freq(phonon: Phonopy):
        """
        Check whether phonon has imaginary frequency.

        Args:
            phonon (Phonopy): Phonopy object which contains phonon spectrum frequency.
        """
        band_dict = phonon.get_band_structure_dict()
        frequencies = np.concatenate(
            [np.array(freq).flatten() for freq in band_dict["frequencies"]], axis=None
        )
        has_imaginary = False
        if np.all(np.array(frequencies) >= -0.299):
            pass
        else:
            print("Warning! Imaginary frequencies found!")
            has_imaginary = True

        return has_imaginary

    def run(self):
        """
        The entrypoint to start the workflow.
        """
        current_path = os.path.abspath(".")
        try:
            # check folder exists
            if not os.path.exists(self.work_dir):
                os.makedirs(self.work_dir)

            os.chdir(self.work_dir)

            try:
                # Generate supercell parameters based on optimized structures
                nrep_second, k_point_mesh = get_supercell_parameters(
                    self.atoms, self.supercell_matrix, self.qpoints_mesh, self.max_atoms
                )
            except Exception as e:
                print("Error whille generating supercell parameters:", e)
                raise

            try:
                # Calculate 2nd force constants
                phonon = self.compute_force_constants(self.atoms, nrep_second)
            except Exception as e:
                print("Error while computing force constants:", e)
                raise

            if self.calc_spec:
                try:
                    # Calculate phonon spectrum
                    self.compute_phonon_spectrum_dos(self.atoms, phonon, k_point_mesh)
                    # check whether has imaginary frequency
                    has_imaginary = self.check_imaginary_freq(phonon)
                except Exception as e:
                    print("Error while computing phonon spectrum and dos:", e)
                    raise
            else:
                has_imaginary = 'Not calculated, set calc_spec True'
                phonon.save(settings={"force_constants": True})

        except Exception as e:
            print("An error occurred during the Phonon workflow:", e)
            raise

        finally:
            os.chdir(current_path)

        return has_imaginary, phonon



import base64
import json
import tempfile
from io import StringIO

import mlflow.pyfunc
import pandas as pd
from ase.io import read as ase_read

from mattersim.cli.applications.moldyn import moldyn
from mattersim.cli.applications.phonon import phonon
from mattersim.cli.applications.relax import relax
from mattersim.cli.applications.singlepoint import singlepoint
from mattersim.forcefield import MatterSimCalculator


artifacts = {
    "checkpoint-1M": "pretrained_models/mattersim-v1.0.0-1M.pth",
    "checkpoint-5M": "pretrained_models/mattersim-v1.0.0-5M.pth",
}


class MatterSimModelWrapper(mlflow.pyfunc.PythonModel):
    def load_context(self, context):
        pass

    def moldyn_wrapper(self, atoms_list, **kwargs):
        kwargs.setdefault("logfile", "md.log")
        kwargs.setdefault("trajectory", "md.traj")

        results = moldyn(atoms_list, **kwargs)

        return results

    def phonon_wrapper(self, atoms_list, **kwargs):
        results = phonon(atoms_list, **kwargs)

        # embed images
        for key in ["phonon_band_plot", "phonon_dos_plot"]:
            for i in range(len(results[key])):
                results[key][i] = base64.b64encode(
                    open(results[key][i], "rb").read()
                ).decode("utf-8")

        return pd.DataFrame(results)

    def relax_wrapper(self, atoms_list, **kwargs):
        results = relax(atoms_list, **kwargs)

        return pd.DataFrame(results)

    def singlepoint_wrapper(self, atoms_list, **kwargs):
        results = singlepoint(atoms_list, **kwargs)

        return pd.DataFrame(results)

    def predict(self, context, model_input, params=None):
        data = json.loads(model_input["data"].item())
        # print(f"Inputs:\n{data}")

        wrappers = {
            "molecular_dynamics": self.moldyn_wrapper,
            "phonon": self.phonon_wrapper,
            "relax_structure": self.relax_wrapper,
            "singlepoint": self.singlepoint_wrapper,
        }

        try:
            wrapper = wrappers[data["workflow"]]
        except KeyError:
            return dict(error="Invalid workflow selected")

        atoms_list = []
        for structure_data in data["structure_data"]:
            atoms_list.extend(ase_read(StringIO(structure_data), format="cif", index=":"))

        mattersim_model = "mattersim-v1.0.0-1m"
        calc = MatterSimCalculator(load_path=mattersim_model)
        for atoms in atoms_list:
            atoms.calc = calc

        with tempfile.TemporaryDirectory() as td:
            data["work_dir"] = td

            df = wrapper(atoms_list, **data)
            output = json.loads(df.to_json(orient="records"))

        # print(f"Output:\n{output}")
        return output


mlflow_pyfunc_model_path = "./mattersim_mlflow_pyfunc"
mlflow.pyfunc.save_model(
    path=mlflow_pyfunc_model_path,
    python_model=MatterSimModelWrapper(),
    artifacts=artifacts,
    conda_env={
        "name": "mattersim-mlflow-env",
        "channels": ["conda-forge"],
        "dependencies": [
            "python=3.9",
            "pip<=24.2",
            "git",
            {
                "pip": [
                    "mlflow==2.19.0",
                    "cloudpickle==3.1.0",
                    "jaraco-collections==5.1.0",
                    "jinja2==3.1.5",
                    "matplotlib==3.9.4",
                    "git+https://github.com/microsoft/mattersim.git@e12460bdc007651716101635c817c33ae9cd81c9",
                    "numpy==1.26.4",
                    "pandas==2.2.3",
                    "phonopy==2.33.4",
                    "platformdirs==4.2.0",
                    "scipy==1.13.1",
                    "torch==2.4.1",
                ],
            },
        ],
    },
)



# -*- coding: utf-8 -*-
from ase import Atoms
from phonopy import Phonopy
from phonopy.structure.atoms import PhonopyAtoms


def get_primitive_cell(atoms: Atoms):
    """
    Get primitive cell from ASE atoms object
    Args:
        atoms (Atoms): ASE atoms object to provide lattice information
    """
    phonopy_atoms = Phonopy(
        to_phonopy_atoms(atoms), primitive_matrix="auto", log_level=2
    )
    primitive = phonopy_atoms.primitive
    atoms = to_ase_atoms(primitive)
    return atoms


def to_phonopy_atoms(atoms: Atoms):
    """
    Transform ASE atoms object to Phonopy object
    Args:
        atoms (Atoms): ASE atoms object to provide lattice informations.
    """
    phonopy_atoms = PhonopyAtoms(
        symbols=atoms.get_chemical_symbols(),
        cell=atoms.get_cell(),
        masses=atoms.get_masses(),
        positions=atoms.get_positions(),
    )
    return phonopy_atoms


def to_ase_atoms(phonopy_atoms):
    """
    Transform Phonopy object to ASE atoms object
    Args:
        phonopy_atoms (Phonopy): Phonopy object to provide lattice informations.
    """
    atoms = Atoms(
        symbols=phonopy_atoms.symbols,
        cell=phonopy_atoms.cell,
        masses=phonopy_atoms.masses,
        positions=phonopy_atoms.positions,
        pbc=True,
    )
    return atoms
